package Make::PurePerl;
use strict;
use Carp qw(croak);
use Sort::Topological qw(toposort);

use File::Path;
use File::Basename;
use File::Find::Rule (); # rule() conflicts otherwise :-))

use vars qw($VERSION);
$VERSION='0.11';

=head1 NAME

Make::PurePerl - make functionality for Perl

=head1 SYNOPSIS

=for example begin

  use Make::PurePerl;

  my $set = Make::PurePerl->new();

  $set->rule( $target, sub { ... }, @sources );
  $set->name( 'cache' )
        ->rule(@targets, sub { ... }, @sources);
  $set->rule( $output, sub { ... }, \'cache' );

  # if you don't like the direction of parameters,
  # try the pipeline syntax:
  $set->pipeline( @sources => sub { ... } => @targets );

  # now kick it off:
  $set->make();

=for example end

=head2 C<new>

Returns a new ruleset.

=cut

sub new {
  my $class = shift;
  my $self = {
    rules => {},
  };
  bless $self, $class;
  $self;
};

=head2 C<rule TARGET CODE SOURCES [ARGS]>

The parameter list for C<rule> is best read as "To create TARGET,
apply CODE to SOURCES".

The optional C<ARGS> must be a hash reference and can contain the rule name
if you don't want to use the C<name> method.

=cut

sub rule {
  my $self = shift;
  my (@target,$code,@source,$args);

  push @target, shift
    while (@_ and ref $_[0] ne 'CODE');
  croak "Need a code reference"
    unless ref $_[0] eq 'CODE';
  $code = shift;
  $args = pop
    if ref $_[-1] eq 'HASH';
  @source = @_;
  $args ||= {};

  $self->add( source => \@source, target => \@target, code => $code, %$args);
};

sub __find_pairs {
  my ($search_template, $result_template) = @_;

  my $base_dir = dirname $search_template;
  my ($template_re);
  ($template_re = basename $search_template) =~ s!%!\(.*\)!;
  $template_re = qr($template_re);

  grep { ! /\bCVS\b/ }
  map {
    m!\Q$base_dir/\E$template_re!
      or croak "Weirdly enough, $_ didn't match /$template_re/";
    my $name;
    my $substitution = $1;
    ($name = $result_template) =~ s!%!$substitution!g;
    $_ => $name
  } (File::Find::Rule->file()->name(qr(^$template_re$))->in($base_dir));
};

sub __file_pairs {
  my ($search_template, $result_template) = @_;
  my %unique = __find_pairs($search_template, $result_template);
  %unique;
};

=head2 C<< auto_rule TARGETSPEC, CODE, SOURCESPEC, PREREQUISITES >>

C<auto_rule> automatically creates rules using a wildcard pattern
like the C<%> expansion in GNU C<make>.

  $set->auto_rule( 'output/%.html', \&make_page, 'input/%.content', @templates );

The above line would add a target C<output/foo.html> for C<input/foo.content>,
and so on, recursing downwards, and ignoring files in any directory called
C<CVS>. Intermediate directories will also be created automatically.

The behaviour of ignoring CVS subdirectories cannot be changed. The later
intention is to have a L<File::Find::Rule> rule for ignored files.

=cut

sub auto_rule {
  my ($self, $target_spec, $code, $source_spec, @other_prerequisites) = @_;

  my %target = __file_pairs($source_spec,$target_spec);

  # Also create rules to create subdirectories if necessary:
  my %target_directories;
  $target_directories{dirname $_} = dirname $target{$_}
    for (keys %target);
  for my $dir (sort keys %target_directories) {
    $self->rule( $target_directories{$dir},
                 sub { my $newdir = shift; unless (-d $newdir) { mkpath $newdir or croak "Couldn't create $newdir: $!" }},
                 $dir ); # needs a (dummy) source element ...
  };

  for my $source (sort keys %target) {
    $self->rule( $target{$source}, $code, $source, @other_prerequisites, dirname $target{$source} );
  };

  $self;
};

=head2 C<pipeline SOURCES CODE TARGET [ARGS]>

The parameter list for C<pipeline> is best read as "Take SOURCES, apply CODE
to create TARGET". Otherwise, it is just syntactic candy around C<rule>, to
allow for left-to-right reading of the parameter list as a pipeline transformation
much like shell syntax:

  $set->pipeline( @sources => sub { ... } => $target );

=cut

sub pipeline {
  my $self = shift;
  my (@source,$code,@target,$args);

  push @source, shift
    while (@_ and ref $_[0] ne 'CODE');
  croak "Need a code reference"
    unless ref $_[0] ne 'CODE';
  $code = shift;
  $args = pop
    if ref $_[-1] eq 'HASH';
  @target = @_;
  $args ||= {};

  $self->add( source => \@source, target => \@target, code => $code, %$args);
};

=head2 C<name NAMES>

Adds the subsequent transformation rules to the meta targets
specified in NAMES. This is syntactic sugar:

  $set->name('cache')
        ->rule('foo1', sub {...}, 'bar1')
        ->rule('foo2', sub {...}, 'bar2');
  $set->rule('baz', sub {...}, \'cache');

is equivalent to

  $set->rule('foo1', sub {...}, 'bar1')
      ->rule('foo2', sub {...}, 'bar2');
  @cache = qw(bar1 bar2);
  $set->rule(@cache, sub {...}, 'baz');

The advantage of C<name> is, that it allows you to delay
the creation of the list up to the latest possible moment
and dosen't clutter your program with it.

C<name> returns a special object that carries the name of
the transformation with it, allowing you to do method chaining.

=cut

sub name {
  my $self = shift;
  Make::PurePerl::NamedRule->new( __DELEGATE => $self, meta_targets => [ @_ ])
};

=head2 C<add %ARGS>

This is the internal method called by C<rule> and C<pipeline> to
add a processing rule.

The keys for the arguments are:

  source       => arrayref of source items
  target       => arrayref of target filenames
  code         => coderef of the transformation to apply
  meta_targets => arrayref of meta-targets this rule also belongs to

=cut

sub add {
  my ($self, %args) = @_;
  my ($meta) = join ":", @{$args{meta_targets} || [""]};
  my (@targets) = @{$args{target}};
  my (@sources) = @{$args{source}};

  # Every source is also a target
  for my $source (@sources) {
    croak "Meta targets are not implemented yet."
      if ref $source;
    $self->target($source);
  };

  for my $target (@targets) {
    croak "Meta targets are not implemented yet."
      if ref $target;
    $self->add_rule(source => $args{source}, code => $args{code}, target => $target);
  };
  $self;
};

sub add_rule {
  my ($self,%args) = @_;
  my $rule = Make::PurePerl::Target::File->new($args{target},$args{code},$args{source});
  push @{$self->target($rule->target)}, $rule;
  $rule;
};

sub target {
  my ($self,$name) = @_;
  $self->{rules}->{$name} ||= [];
};

sub targets { keys %{ +shift->{rules} }; };

=head2 C<ordered_targets>

Returns all targets ordered in a fashion that, if the rules for each
target are executed one after another, all targets will be
up to date. Note that no check is made whether the code attached to a rule
needs to be executed or not.

If you see the set of make rules as a partially ordered set, this function
returns the set of rules in a partial order. The method used is to convert
the partial order to a total order by comparing the addresses of
code blocks in the case of incomparability.

=cut

sub ordered_targets {
  my $self = shift;
  my %children;

  for my $target ($self->targets) {
    my %uniq;
    $uniq{$_} = 1
      for map { $_->target } @{$self->target($target)};
    delete $uniq{$target};
    $children{$target} = [ keys %uniq ];
  };

  return toposort( sub{ @{$children{ +shift }} }, [keys %children]);
};

=head2 C<make>

Applies all rules in a correct order and executes the code
for the rules associated with them where at least one
prerequisite is newer than the target. Makes stuff happen.

=cut

sub apply {
  my ($self,$callback) = @_;
  $callback ||= sub {};
  for my $target ($self->ordered_targets) {
    my @rules = @{$self->target($target)};
    for my $rule (@rules) {
      $callback->($target);
      $rule->must_run() and $rule->run();
    };
  };
};

=head2 C<force>

Applies all rules in a correct order and executes the code
associated with them. Makes stuff happen.

=cut

sub force {
  my ($self,$callback) = @_;
  $callback ||= sub {};
  for my $target ($self->ordered_targets) {
    my @rules = @{$self->target($target)};
    for my $rule (@rules) {
      $callback->($target);
      $rule->run();
    };
  };
};

package Make::PurePerl::Target::File;
use strict;
use Carp qw(croak);
use base 'Class::Accessor';

__PACKAGE__->mk_accessors(qw(target source code));

=head1 NAME

Make::PurePerl::Target::File - File target

=cut

sub new {
  my ($class,$target,$code,$source) = @_;
  my $self = {
    source => [ @{$source} ],
    target => $target,
    code   => $code,
  };
  bless $self,$class;
  $self;
};

sub newer_files {
  my ($self) = @_;
  return @{$self->source} unless -f $self->target;

  my $mtime = -M $self->target;
  croak sprintf( "Couldn't find mtime for %s: $!", $self->target )
    unless defined $mtime;

  return grep { -M $_ < $mtime } (@{$self->source});
};

sub must_run { scalar (shift->newer_files); };

sub run {
  my ($self) = @_;
  $self->code->($self->target, @{$self->source});
};

package Make::PurePerl::NamedRule;
use strict;
use Carp qw(croak);

=head1 NAME

Make::PurePerl::NamedRule - Very simple container for named rules

=cut

sub new {
  my ($class,%args) = @_;
  my $delegate = delete $args{__DELEGATE};
  croak "__DELEGATE key is missing! Don't know who to delegate to!" unless $delegate;

  my $self = {
    delegate => $delegate,
    args => \%args,
  };
  bless $self,$class;

  $self;
};

# explicitly state stuff we don't want to AUTOLOAD
sub DESTROY {};

sub AUTOLOAD {
  my ($self) = shift;
  our $AUTOLOAD =~ /^.*::([^:]+)/
    or croak "Can't find a method name in $AUTOLOAD";
  my $method = $1;
  my $rulename = join ",", @{$self->{args}->{meta_targets}};
  croak "Can't add $method name: The rule names are already set to $rulename"
    if ref $_[-1] eq 'HASH';
  $self->{delegate}->$method(@_,$self->{args});
  $self;
};

1;

__END__

=head1 AUTHOR

Max Maischein, E<lt> corion at cpan.org E<gt>

=head1 SEE ALSO

L<ppmake>, L<Algorithm::Dependencies>, L<Tree::DAG_Node>.

=cut
